;; Generated package description from ediprolog.el  -*- no-byte-compile: t -*-
(define-package "ediprolog" "2.2" "Emacs Does Interactive Prolog" 'nil :commit "cfcdf9e42821d246b7fbc84877aa4ecacc184a1c" :authors '(("Markus Triska" . "triska@metalevel.at")) :maintainer '("Markus Triska" . "triska@metalevel.at") :keywords '("languages" "processes") :url "https://www.metalevel.at/ediprolog/")
